/*
 * libstorage
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef __STORAGE_EXTERNAL_DBUS_H__
#define __STORAGE_EXTERNAL_DBUS_H__

#include <stdbool.h>
#include <glib.h>
#include <gio/gio.h>
#include "list.h"

#define STORAGE_EXT_BUS_NAME              "org.tizen.system.storage"
#define STORAGE_EXT_PATH_MANAGER          "/Org/Tizen/System/Storage/Block/Manager"
#define STORAGE_EXT_IFACE_MANAGER         STORAGE_EXT_BUS_NAME".BlockManager"
#define STORAGE_EXT_PATH_STORAGE          "/Org/Tizen/System/Storage/Storage"
#define STORAGE_EXT_IFACE_STORAGE         STORAGE_EXT_BUS_NAME".storage"

enum mount_state {
	STORAGE_EXT_UNMOUNTED,
	STORAGE_EXT_MOUNTED,
};

enum storage_ext_state {
	STORAGE_EXT_REMOVED,
	STORAGE_EXT_ADDED,
	STORAGE_EXT_CHANGED,
	STORAGE_EXT_BLOCKED,
};

enum storage_ext_type {
	STORAGE_EXT_SCSI,
	STORAGE_EXT_MMC,
	STORAGE_EXT_MMC_EXTENDED_INTERNAL,
};

enum storage_ext_flags {
	FLAG_NONE        = 0,
	UNMOUNT_UNSAFE   = 1 << 0,
	FS_BROKEN        = 1 << 1,
	FS_EMPTY         = 1 << 2,
	FS_NOT_SUPPORTED = 1 << 3,
	MOUNT_READONLY   = 1 << 4,
};

typedef struct _storage_ext_device {
	enum storage_ext_type type;
	char *devnode;
	char *syspath;
	char *fs_usage;
	char *fs_type;
	char *fs_version;
	char *fs_uuid;
	bool readonly;
	char *mount_point;
	enum mount_state state;
	bool primary;   /* the first partition */
	int flags;
	int storage_id;
} storage_ext_device;

/* it's for 32bit file offset */
struct statvfs_32 {
	unsigned long int f_bsize;
	unsigned long int f_frsize;
	unsigned long int f_blocks;
	unsigned long int f_bfree;
	unsigned long int f_bavail;
	unsigned long int f_files;
	unsigned long int f_ffree;
	unsigned long int f_favail;
	unsigned long int f_fsid;
#ifdef _STATVFSBUF_F_UNUSED
	int __f_unused;
#endif
	unsigned long int f_flag;
	unsigned long int f_namemax;
	int __f_spare[6];
};

typedef int (*storage_ext_changed_cb)(storage_ext_device *dev, enum storage_ext_state state, void *data);

int storage_ext_is_supported(void);
void storage_ext_release_device(storage_ext_device **dev);
void storage_ext_release_list(dd_list **list);
int storage_ext_get_list(dd_list **list);
int storage_ext_get_statvfs(char *path, struct statvfs_32 *buf);
int storage_ext_get_statvfs_size64(char *path, struct statvfs *buf);

int storage_ext_register_device_change(storage_ext_changed_cb func, void *data);
void storage_ext_unregister_device_change(storage_ext_changed_cb func);

int storage_ext_get_device_info(int storage_id, storage_ext_device *info);
int storage_ext_get_storage_level(const char *path, char **level);

typedef void (*dbus_pending_cb)(GVariant *var, void *user_data, GError *err);
int dbus_method_async_with_reply_var(const char *dest, const char *path,
		const char *iface, const char *method, GVariant *param,
		dbus_pending_cb cb, int timeout, void *data);
/* storage-internal.c */
GVariant *dbus_method_call_sync(const gchar *dest, const gchar *path,
		const gchar *iface, const gchar *method, GVariant *param);

#endif /* __STORAGE_EXTERNAL_DBUS_H__ */
